import psutil
from typing import Callable, List, Tuple, Optional
from wasabi import msg
import time
from .tokenizer import TokenList


def get_cpu_memory(tag: str):
    """获取函数运行时的cpu内存

    Args:
        tag (str): 标签
    """
    def out_wrapper(fn: Callable):
        def wrapper(*args, **kwargs):
            p = psutil.Process()
            cpu_memory = p.memory_info().rss
            outputs = fn(*args, **kwargs)
            cpu_memory = p.memory_info().rss - cpu_memory
            msg.good(f"{tag} cpu memory: {cpu_memory / 1024 / 1024}MB")
            return outputs
        return wrapper
    return out_wrapper

def get_spent_time(tag: str):
    """获取函数运行时的时间

    Args:
        tag (str): 标签
    """
    def out_wrapper(fn: Callable):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            outputs = fn(*args, **kwargs)
            end_time = time.time()
            msg.good(f"{tag} spent time: {end_time - start_time}")
            return outputs
        return wrapper
    return out_wrapper



def get_ents(tags: List[str]) -> List[Tuple[int, int, str]]:
    """从序列标签中提取实体

    Args:
        tags (List[str]): 序列标签.

    Returns:
        List[Tuple[int, int, str]]: 实体列表.例如, [(2, 6, 'PER')]
    """
    entities = []
    entity = []
    for i, tag in enumerate(tags):
        if tag.startswith('B-'):
            if entity:
                entities.append(tuple(entity))
            entity = [i, i + 1, tag.split('-')[1]]
        elif tag.startswith('I-'):
            if entity:
                if entity[2] == tag.split('-')[1]:
                    entity[1] = i + 1
        else:
            if entity:
                entities.append(tuple(entity))
            entity = []
    if len(entity) == 3:
        entities.append(tuple(entity))
    return entities
        
    
def ner_post_process(labels: List[str], tokens: Optional[TokenList] = None) -> List[Tuple[int, int, str, str]]:
    """对ner模型预测的token标签进行后处理，得到最终的实体

    Args:
        labels (List[str]): 模型预测的序列标签.
        text (str): 原始文本.

    Returns:
        List[Tuple[int, int, str, str]]: 实体列表.例如, [(2, 6, '明天上午', 'DateChanged')]
    """
    if tokens:
        assert len(labels) == len(tokens)
    ents = get_ents(labels)
    if not tokens:
        return ents
    results = []
    for ent in ents:
        start, end, label = ent
        start_char = tokens.token_to_char(start)[0]
        end_char = tokens.token_to_char(end - 1)[-1] + 1
        results.append((start_char, end_char, tokens.processed_text[start_char:end_char], label))
    return results


from typing import Any, List, Tuple, Literal
from onnxruntime import InferenceSession, SessionOptions
from fasttext import load_model
from pathlib import Path
from .utils import ner_post_process
import json
import numpy as np
from .tokenizer import Tokenizer
from pydantic import validate_arguments



class FasttextInference():
    def __init__(self, model_path: str):
        super().__init__()
        if isinstance(model_path, str):
            self.model = load_model(model_path)
        if isinstance(model_path, Path):
            self.model = load_model(str(model_path))
    
    def __call__(self, query: str) -> Tuple[str, float]:
        """fasttext 模型推理

        Args:
            query (str): 原始问题文本
        """
        query = ' '.join(list(query))
        result = self.model.predict(query)
        label = result[0][0]
        score = result[1][0]
        return label, round(score, 4)
    
    
class NERInferenceLocal():
    """ner模型推理, 返回格式如下 (2, 6, '明天上午', 'DateChanged')
    """
    def __init__(self, 
                 embedding_model_path: str,
                 classifier_model_path: str,
                 tokenizer: Tokenizer,
                 label_path: str):
        super().__init__()
        so = SessionOptions()
        so.log_severity_level = 3
        if isinstance(classifier_model_path, str):
            self.classifier_model = InferenceSession(classifier_model_path, so)
        if isinstance(embedding_model_path, Path):
            self.classifier_model = InferenceSession(str(classifier_model_path), so)
        
        self.tokenizer = tokenizer
        
        with open(label_path, 'r') as f:
            self.id2label = json.load(f)
            
        self.embedding_model = InferenceSession(embedding_model_path, so)
    
    def __call__(self, query: str) -> List[Tuple[int, int, str, str]]:
        """ner 模型推理

        Args:
            query (str): 原始问题文本
        """
        tokens = self.tokenizer(query, padding=True)[0]
        embeddings = self.get_embeddings(ids=tokens.ids)
        labels = self.get_labels(embeddings)
        
        # 后处理
        ents = ner_post_process(labels, tokens=tokens)
        return ents
    
    
    def get_embeddings(self, ids: List[int]) -> np.ndarray:
        """获取embedding

        Args:
            query (str): 原始问题文本
        """
        input_ids = np.array(ids, dtype=np.int64).reshape(1, -1)
        # 获取embedding
        embeddings = self.embedding_model.run(None, {'input': input_ids})[0]
        embeddings = np.array(embeddings, dtype=np.float32)
        return embeddings
    
    
    def get_labels(self, embeddings: np.ndarray) -> List[str]:
        """获取标签

        Args:
            embeddings (np.ndarray): embedding
        """
        label_ids = self.classifier_model.run(None, {'input': embeddings})[0][0]
        # 兼容tf版本转换的onnx模型
        if len(label_ids.shape) == 2:
            label_ids = np.argmax(label_ids, axis=-1)
            
        labels = [self.id2label[str(i)] for i in label_ids]
        return labels
    
    
    
class NERInferenceCloud():
    """ner模型推理, 返回格式如下 (2, 6, '明天上午', 'DateChanged')
    """
    def __init__(self, 
                 model_path: str,
                 tokenizer: Tokenizer,
                 label_path: str):
        super().__init__()
        
        so = SessionOptions()
        so.log_severity_level = 3
        
        self.model = InferenceSession(str(model_path), so)
        
        self.tokenizer = tokenizer
        
        with open(label_path, 'r') as f:
            self.id2label = json.load(f)
    
    
    @validate_arguments
    def __call__(self, query: str) -> List[Tuple[int, int, str, str]]:
        """ner 模型推理

        Args:
            query (str): 原始问题文本
        """
        tokens = self.tokenizer([query], padding=True)[0]
        
        label_ids = self.model.run(None, {'input': np.array([tokens.ids], dtype=np.int64)})[0][0]
        # 兼容tf版本转换的onnx模型
        if len(label_ids.shape) == 2:
            label_ids = np.argmax(label_ids, axis=-1)
            
        labels = [self.id2label[str(i)] for i in label_ids]
        
        # 后处理
        ents = ner_post_process(labels, tokens=tokens)
        
        return ents